# pylint: skip-file
import unittest

from sqlalchemy import Column
from sqlalchemy import String

from unimatrix.ext.orm import Relation
from unimatrix.ext.orm.testing import TestDatabaseManager


class TestDatabaseManagerTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.manager = TestDatabaseManager([Relation])
        cls.manager.on_setup_class({
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ":memory:"
        })

    def test_setup_and_teardown_have_clean_state(self):
        self.manager.on_setup(self)
        try:
            dao = TestModel(foo='bar')
            self.session.add(dao)
            self.session.commit()

            q = self.session.query(TestModel)\
                .filter(TestModel.foo=='bar')\
                .exists()
            self.assertTrue(self.session.query(q).scalar())
        finally:
            self.manager.on_teardown()

        try:
            session = self.manager.get_session()
            q = session.query(TestModel)\
                .filter(TestModel.foo=='bar')\
                .exists()
            self.assertFalse(session.query(q).scalar())
        finally:
            session.close()


class TestModel(Relation):
    __tablename__ = 'testmodel'

    foo = Column(String, primary_key=True)



"""Declares various functions to assist in testing with :mod:`sqlalchemy`."""
import asyncio
import copy
import os
import unittest

import ioc
import sqlalchemy
from ioc.loader import import_symbol
from sqlalchemy import create_engine
from sqlalchemy import event
from sqlalchemy import select
from sqlalchemy.orm import sessionmaker

import unimatrix.ext.orm.conf
from unimatrix.ext.orm import Relation
from unimatrix.ext.orm import async_session
from unimatrix.ext.orm import create_engine
from unimatrix.ext.orm import destroy_engines
from unimatrix.ext.orm.conf import load_config


class AsyncSessionTestCase(unittest.TestCase):
    ioc_session_factory_key = None
    target_metadata = []

    @property
    def context_managers(self):
        return list(dict.values(self.transactions))\
            + list(dict.values(self.sessions))

    def run_sync(self, f, *args, **kwargs):
        return self.loop.run_until_complete(f(*args, **kwargs))

    def setUp(self):
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        # If metadata is specified, create it using the default
        # connection.
        if self.target_metadata:
            engine = create_engine(use_async=True)
            for metadata in self.target_metadata:
                tx = engine.begin()
                connection = self.run_sync(tx.__aenter__)
                self.run_sync(connection.run_sync, metadata.create_all)

        # Create session objects for all configured connections.
        self.sessions = {}
        self.transactions = {}
        for name in self.get_connections():
            Session = async_session(name)
            self.sessions[name] = session =\
                self.run_sync(Session.__aenter__)

            # Start a transaction and roll it back in the tearDown()
            # method.
            self.transactions[name] = tx = session.begin_nested()
            self.run_sync(tx.__aenter__)
            self.run_sync(session.execute, select(1))

        if self.ioc_session_factory_key:
            ioc.provide(self.ioc_session_factory_key,
                self.get_session, force=True)

    def tearDown(self, destroy=True):
        for name, tx in dict.items(self.transactions):
            self.run_sync(tx.rollback)
        for ctx in dict.values(self.sessions):
            self.run_sync(ctx.__aexit__, None, None, None)
        if destroy:
            destroy_engines(loop=self.loop)

    def execute_query(self, session, q, name='self'):
        """Execute query `q` and return the awaited result."""
        return self.run_sync(session.execute, q)

    def get_connections(self):
        """Return the list holding all connection names."""
        return list(dict.keys(unimatrix.ext.orm.conf.DATABASES))

    def get_session(self, name='self'):
        """Returns the overridden session factory for `name`."""
        return self.sessions[name]

    async def create_all(self, engine, metadata):
        async with engine.begin() as connection:
            metadata.create_all(connection)


class RepositoryTestCase(AsyncSessionTestCase):
    """A system test case for repository implementations that use a
    relational database.
    """
    repo_class = None
    ioc_session_factory_key = 'AsyncSessionFactory'

    def setUp(self):
        if self.repo_class is None:
            self.fail(
                "%s must specify a repository class on the "
                "`repo_class` attribute" % type(self).__name__
            )
        super().setUp()
        self.repository = self.repo_class().with_context()
        self.run_sync(self.repository.__aenter__)

    def tearDown(self):
        self.run_sync(self.repository.__aexit__, None, None, None)
        super().tearDown()


class TestCase(unittest.TestCase):
    """A :class:`unittest.TestCase` implementation that ensures a clean
    database state between tests.
    """
    default_bases = [Relation]

    #: Additional base classes for SQLAlchemy models.
    bases = None

    @classmethod
    def setUpClass(cls):
        cls.manager = TestDatabaseManager(
            self.default_bases + list(self.bases or []))
        cls.manager.on_setup_class({
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ":memory:"
        })

    def setUp(self):
        """Ensures that all connections, transactions and sessions are
        started.
        """
        self.manager.on_setup(self)

    def tearDown(self):
        """Rolls back all transactions and closes the connections."""
        self.manager.on_teardown()


class TestDatabaseManager:
    """Manages the state of the database(s) during automated tests."""

    def __init__(self, bases):
        self.bases = bases
        self.databases = {}
        self.engines = {}
        self.connections = {}
        self.transactions = {}
        self.sessions = {}
        self.session = None

    def get_session(self, name='self'):
        """Return a session for the named connection."""
        return sessionmaker(bind=self.engines[name])()

    def on_setup_class(self, databases=None):
        """Prepare the database connections to commence testing."""
        # Get the database from the environment and secrets directory, and
        # configure a default connection if one is not specified.
        self.databases = load_config(env=copy.deepcopy(os.environ))
        if not self.databases:
            self.databases = load_config(env={
                'DB_ENGINE': "sqlite",
                'DB_NAME': ":memory:"
            })

        # Loop over the discovered database connections to create the engine
        # objects.
        for name, opts in dict.items(self.databases):
            self.engines[name] = engine = opts.as_engine(
                sqlalchemy.create_engine, echo=True)
            if name == 'self':
                for Base in self.bases:
                    Base.metadata.create_all(engine)

    def on_setup(self, testcase):
        """Ensures that all sessions for the configured databases have started
        a transaction.
        """

        # TODO: Use the default engine to create the tables. Multi-database
        # support is not fully implemented like Django, but this is good
        # enough for now.
        for name in dict.keys(self.databases):
            engine = self.engines[name]
            self.connections[name] = connection = engine.connect()
            self.transactions[name] = tx = connection.begin()
            self.sessions[name] = session = sessionmaker(bind=connection)()
            if name == 'self':
                testcase.session = session
            #session.begin_nested()

            #@event.listens_for(session, "after_transaction_end")
            #def restart_savepoint(session, tx):
            #    if tx.nested and not tx._parent.nested:
            #        session.begin_nested()

    def on_teardown(self):
        """Rollback all transactions and dispose the engines."""
        for session in dict.values(self.sessions):
            session.close()
        for tx in dict.values(self.transactions):
            tx.rollback()
        for connection in dict.values(self.connections):
            connection.close()



# pylint: skip-file
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy.ext.declarative import declarative_base


Base = declarative_base()


class Foo(Base):
    __tablename__ = 'foo'

    bar = Column(Integer, primary_key=True)



# pylint: skip-file
import unittest

from ..conf import DatabaseConnection


class DatasourceConstructionTestCase(unittest.TestCase):
    
    def test_postgresql_minimal_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://localhost:5432")
    
    def test_postgresql_minimal_host_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'host': "rdbms"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://rdbms:5432")
    
    def test_postgresql_minimal_port_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'port': "1000"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://localhost:1000")
    
    def test_postgresql_minimal_name_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'name': "rdbms"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://localhost:5432/rdbms")
    
    def test_postgresql_username_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'username': "rdbms"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://rdbms@localhost:5432")
    
    def test_postgresql_username_password_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'username': "rdbms",
            'password': "rdbms"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://rdbms:rdbms@localhost:5432")
    
    def test_postgresql_full_sync(self):
        conn = DatabaseConnection({
            'engine': "postgresql",
            'host': "rdbms",
            'port': "1111",
            'name': "rdbms",
            'username': "rdbms",
            'password': "rdbms"
        })
        self.assertEqual(conn.as_dsn(),
            "postgresql+psycopg2://rdbms:rdbms@rdbms:1111/rdbms")



"""Declares the base class for all models in the Unimatrix framework."""
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import DeclarativeMeta


class RelationCollection(dict):
    """Registry for all mapped classes that inherit from :class:`Relation`."""
    __name__ = 'RelationCollection'


# All classes that are mapped in the Unimatrix packages can be found in this
# dictionary.
RELATIONS = RelationCollection()


class RelationMeta(DeclarativeMeta):
    pass


class RelationBase:
    pass


Relation = declarative_base(
    class_registry=RELATIONS,
    cls=RelationBase,
    metaclass=RelationMeta,
    name='Relation'
)



"""Declares :mod:`pytest` fixtures to assist in testing."""
import os

import ioc.loader
import pytest
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.runtime.environment import EnvironmentContext
from sqlalchemy import engine_from_config
from sqlalchemy import pool
from sqlalchemy.orm import sessionmaker
from unimatrix.conf import settings

from .. import conf
from .. import create_engine
from .. import destroy_engines


def requires_rdbms(test, *args, **kwargs):
    """Class or function decorator that ensures that the testing database
    is configured during a test session.
    """
    return pytest.mark.usefixtures('setup_databases')(test)


def _setup_sqlite(name, connection):
    connection['name'] = ":memory:"


def _setup_postgres(name, connection):
    db_name = connection['name']
    test_db_name = f"test_{db_name}"

    # Connect to the postgres database (which is assumed to exist) so we
    # can issue the CREATE DATABASE statement.
    engine = create_engine(name)
    session = sessionmaker(bind=engine)()
    session.connection().connection.set_isolation_level(0)
    session.execute(f"DROP DATABASE IF EXISTS {test_db_name}")
    session.execute(f"CREATE DATABASE {test_db_name}")
    connection['name'] = test_db_name


_setup_functions = {
    'postgresql': _setup_postgres,
    'sqlite': _setup_sqlite
}


@pytest.fixture(scope='session')
def setup_databases():
    """Lookup all database connections and reconfigure the internal state
    of the :mod:`unimatrix.ext.orm` package. Run the migrations using
    alembic.
    """
    test_databases = {}
    for name in dict.keys(conf.DATABASES):
        connection = conf.DATABASES[name]
        _setup_functions[ connection['engine'] ](name, connection)

    # Ensure that all engines are disposed and sessions closed. All new
    # engines and sessions connect to the test database.
    if not os.path.exists('alembic.ini'):
        return
    self = conf.DATABASES['self']
    target_metadata = ioc.loader.import_symbol(settings.SQL_METADATA)
    dsn = self.as_dsn()
    config = Config('alembic.ini')
    config.set_main_option("sqlalchemy.url", dsn)
    script = ScriptDirectory.from_config(config)
    context = EnvironmentContext(config, script)
    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    def do_upgrade(revision, context):
        return script._upgrade_revs(script.get_heads(), revision)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            fn=do_upgrade
        )

        with context.begin_transaction():
            context.run_migrations()

    destroy_engines()



"""Declares :class:`NamedQuery`."""
import marshmallow.exceptions
from marshmallow.fields import *
from marshmallow import Schema
from sqlalchemy.orm.exc import NoResultFound
from unimatrix.ext.model.exc import ValidationError


FORBIDDEN_ATTRS = ['build', 'execute', 'schema_context']


class NamedQueryParameter:

    def __init__(self, name):
        self.name = name

    def __get__(self, obj, cls=None):
        return obj._params[self.name]


class NamedQueryMeta(type):

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        if name == 'NamedQuery':
            return super_new(cls, name, bases, attrs)

        # Construct a marshmallow.Schema class based on the fields found
        # on the NamedQuery subclass.
        fields = {}
        for attname in list(dict.keys(attrs)):
            if not isinstance(attrs[attname], Field):
                continue
            if attname in FORBIDDEN_ATTRS:
                raise ValueError("Field name '%s' is not allowed" % attname)
            fields[attname] = attrs.pop(attname)
            attrs[attname] = NamedQueryParameter(attname)

        attrs['schema_class'] = type("%sSchema" % name, (Schema,), fields)

        return super_new(cls, name, bases, attrs)


class NamedQuery(metaclass=NamedQueryMeta):
    """Represents a named database query."""
    declarative = False

    def __init__(self, *args, **kwargs):
        schema = self.schema_class()
        try:
            self._params = schema.load(kwargs)
        except marshmallow.exceptions.ValidationError as e:
            raise ValidationError(spec=e.normalized_messages())
        self._query = None

    def build(self):
        """Builds the query."""
        raise NotImplementedError

    def process_result(self, result):
        """Hook to process the query results."""
        return result

    def _build(self):
        self._query = self.build()
        return self._query

    def _process_result(self, result):
        return self.process_result(result)

    async def run(self, session):
        """Run the query."""
        return Result(self, await self.execute(session))

    async def execute(self, session):
        """Executes the query and returns the result."""
        return self._process_result(await session.execute(self._build()))


class BaseDeclarativeQuery(NamedQuery):
    """A :class:`~unimatrix.ext.orm.query.Query` implementation that
    process the result as an SQLAlchemy declarative model.
    """

    def process_result(self, result):
        return result.scalars()


class DeclarativeCollectionQuery(BaseDeclarativeQuery):
    """Like :class:`BaseDeclarativeQuery`, but expects a collection."""

    def process_result(self, result):
        return result.scalars()


class DeclarativeQuery(BaseDeclarativeQuery):
    """Like :class:`BaseDeclarativeQuery`, but expects a single entity
    to be returned.
    """

    #: Indicates if the result may be ``None``.
    allow_none = False

    def process_result(self, result):
        result = super().process_result(result)
        return result.one() if not self.allow_none else result.first()


class Get(NamedQuery):
    """A :class:`Query` implementation that results a single row."""
    allow_none = False

    def process_result(self, result):
        result = result.fetchone()
        if result is None and not self.allow_none:
            raise NoResultFound
        return result


class Result:
    """The result of an executed :class:`Query`."""
    not_provided = object()

    def __init__(self, query, result):
        self.__query = query
        self.__result = result

    def __getattr__(self, attname):
        return getattr(self.__result, attname)

    def __iter__(self):
        return iter(self.__result)


Query = NamedQuery



# pylint: skip-file
import asyncio
import unittest

from sqlalchemy import select
from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import inspect
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.exc import NoResultFound

from .. import query


Base = declarative_base()


class NamedQueryIntegrationTestCase(unittest.TestCase):

    def setUp(self):
        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
        self.engine = create_async_engine('sqlite://')
        self.session_factory = AsyncSession(self.engine)
        self.session = self.run_sync(self.session_factory.__aenter__)
        self.run_sync(self.create_metadata)

    def tearDown(self):
        self.run_sync(self.session.__aexit__, None, None, None)
        self.run_sync(self.engine.dispose)

    async def create_metadata(self):
        async with self.engine.begin() as connection:
            await connection.run_sync(Base.metadata.create_all)

        self.session.add(FooModel(value='foo'))
        self.session.add(FooModel(value='bar'))
        self.session.add(FooModel(value='baz'))
        self.session.add(BarModel(value='foo'))
        await self.session.commit()

    def run_sync(self, f, *args, **kwargs):
        return self.loop.run_until_complete(f(*args, **kwargs))

    def test_execute_query_declarative_one(self):
        q = GetFooByValue(value='bar')
        result = self.run_sync(q.run, self.session)
        self.assertEqual(result.value, 'bar')
    
    def test_execute_query_declarative_list(self):
        q = ListFoo()
        result = self.run_sync(q.run, self.session)
        objects = list(result)
        self.assertEqual(objects[0].value, 'foo')
        self.assertEqual(objects[1].value, 'bar')
        self.assertEqual(objects[2].value, 'baz')

    def test_get_column_from_single_query_result(self):
        q = GetFooColumnName(value='bar')
        result = self.run_sync(q.run, self.session)
        self.assertEqual(result.other_value, None)

    def test_get_column_from_single_query_result_no_match(self):
        q = GetFooColumnName(value='baadadr')
        with self.assertRaises(NoResultFound):
            result = self.run_sync(q.run, self.session)

    def test_joined_declarative_columns(self):
        q = JoinedDeclarativeQuery(value='foo')
        result = self.run_sync(q.run, self.session)
        self.assertEqual(result.foo, 'foo')
        self.assertEqual(result.bar, 'foo')



class FooModel(Base):
    __tablename__ = 'foo'
    value = Column(String, primary_key=True)
    other_value = Column(String)


class BarModel(Base):
    __tablename__ = 'bar'
    value = Column(String, primary_key=True)


class GetFooByValue(query.DeclarativeQuery):
    value = query.String()

    def build(self):
        return select(FooModel).where(FooModel.value==self.value)


class GetFooColumnName(query.Get):
    value = query.String()

    def build(self):
        t = inspect(FooModel)
        return select(t.c.other_value).where(t.c.value==self.value)


class ListFoo(query.DeclarativeCollectionQuery):

    def build(self):
        return select(FooModel)


class JoinedDeclarativeQuery(query.Get):
    value = query.String()

    def build(self):
        R1 = inspect(FooModel)
        R2 = inspect(BarModel)
        return select([R1.c.value.label('foo'), R2.c.value.label('bar')])\
            .join(R2, R1.c.value==R2.c.value)\
            .where(R1.c.value==R2.c.value)



# pylint: skip-file
import asyncio
import unittest

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import select
from sqlalchemy.ext.declarative import declarative_base

from .. import create_engine
from .. import destroy_engines
from .. import async_session
from ..conf import load_config


class AsyncSessionTestCase(unittest.TestCase):

    def setUp(self):
        self.engine = create_engine(
            use_async=True,
            databases=load_config({
                'DB_ENGINE': 'sqlite',
                'DB_NAME': ":memory:"
            })
        )
        self.session = async_session()

    def tearDown(self):
        destroy_engines()

    def test_async_create_metadata(self):
        async def f():
            async with self.engine.begin() as engine:
                await engine.run_sync(Base.metadata.create_all)

            async with self.session as session:
                session.add(TestModel(id=1))

                q = select(TestModel).filter(TestModel.id==1)
                result = await session.execute(q)
                dao = result.scalars().first()
                self.assertEqual(dao.id, 1)

        asyncio.run(f())


Base = declarative_base()

class TestModel(Base):
    __test__ = False
    __tablename__ = 'testmodel'

    id = Column(Integer, primary_key=True)



"""Declares :class:`RelationalDatabaseRepository`."""
import ioc


class RelationalDatabaseRepository:
    """A repository implementation for use with :mod:`sqlalchemy`."""
    session_factory = 'AsyncSessionFactory'

    async def query(self, q):
        """Run a :class:`~unimatrix.ext.orm.query.Query` and return
        the result(s).
        """
        return await q.run(self.session)

    async def __aenter__(self):
        if getattr(self, 'session', None) is None:
            if isinstance(self.session_factory, str):
                self.session = ioc.require(self.session_factory)()
            elif callable(self.session_factory):
                self.session = self.session_factory()
            else:
                raise NotImplementedError(
                    'Provide an inversion-of-control key as the session_factory'
                    ' attribute or implement it as a method.'
                )

        self._manages_session = not self.session.sync_session.in_transaction()

        # Determine if there is a transaction running. If a transaction is
        # running, start a savepoint, else begin a new one.
        begin = self.session.begin
        if not self._manages_session:
            begin = self.session.begin_nested
        self.tx = begin()
        await self.tx.__aenter__()
        return self

    async def __aexit__(self, cls, exc, tb):
        try:
            await self.tx.__aexit__(cls, exc, tb)
        finally:
            if self._manages_session:
                await self.session.close()




"""Specifies the module-level command line interface."""
import argparse

from unimatrix.ext import orm


parser = argparse.ArgumentParser("Database maintenance and debugging")
parser.add_argument('subcommand', help="specify the command to run.")


def main(args):
    if args.subcommand != 'awaitservices':
        raise NotImplementedError
    orm.awaitservices()


if __name__ == '__main__':
    main(parser.parse_args())



# pylint: skip-file
from sqlalchemy import select

from unimatrix.ext.orm import async_session
from unimatrix.ext.orm import destroy_engines
from unimatrix.ext.orm.testing import AsyncSessionTestCase
from .orm import Foo


class TestAsyncSessionTestCase(AsyncSessionTestCase):
    target_metadata = [Foo.metadata]

    def test_no_data_is_persisted(self):
        session = self.get_session()
        session.add(Foo())
        
        result = self.execute_query(session, select(Foo))
        self.assertEqual(len(list(result)), 1)

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown(destroy=False)
        try:
            session = async_session()
            result = self.execute_query(session, select(Foo))
            self.assertEqual(len(list(result)), 0)
        finally:
            destroy_engines()



"""The :mod:`unimatrix.ext.orm` package provides common ORM functionality
for Unimatrix packages.
"""
import asyncio
import importlib
import time
from concurrent.futures import ThreadPoolExecutor

import sqlalchemy
import sqlalchemy.ext.declarative
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.ext.asyncio.engine import AsyncEngine
from sqlalchemy.sql.schema import MetaData
from sqlalchemy.orm import sessionmaker

from .base import Relation
from .conf import DatabaseConnection
from .conf import load_config
from .conf import DATABASES
from .repository import RelationalDatabaseRepository


ENGINES = {}
ASYNC_ENGINES = {}


def awaitservices(max_retries=60, interval=1, databases=None):
    """Block until all database connections are available.

    Args:
        max_retries (int): the maximum number of retries. Defaults to ``60``.
        interval (int): the interval between retries, in seconds. Defaults
            to ``1``.

    Returns:
        None

    Raises:
        :exc:`RuntimeError`
    """
    databases = databases or DATABASES
    with ThreadPoolExecutor(max_workers=len(databases)) as executor:
        ready = []
        attempts = 0
        while True:
            if set(ready) == set(dict.keys(databases)):
                break
            if attempts >= max_retries:
                raise RuntimeError(
                    "Connections not ready after %s retries." % attempts)
            futures = executor.map(
                DatabaseConnection.wait,
                [databases[name]
                    for name in dict.keys(databases)
                    if name not in ready
                ]
            )
            for result in futures:
                if result is not None:
                    ready.append(result.name)

            attempts += 1
            time.sleep(interval)


def declarative_base(cls=None, class_factory=None, settings=None, apps=None):
    """Like :func:`sqlalchemy.ext.declarative.declarative_base()`, but
    returns a tuple containing the metadata objects for all applications
    that were specified in :attr:`unimatrix.conf.settings.INSTALLED_APPS`.

    Each application must have a submodule/subpackage named ``orm``, which
    exposed a ``metadata`` attribute.

    The list of installed apps may also be provided using the `apps`
    parameter.

    The resulting list of metadata objects is used to create Alembic
    migration scripts.

    Optionally, the `cls` argument may be provided to use a pre-existing
    declarative base class instead. This allows the :func:`declarative_base`
    function to be used with legacy models.

    By default, :func:`sqlalchemy.ext.declarative.declarative_base()` is
    used to create the base class, but another factory may be specified using
    the `class_factory` argument.
    """
    class_factory = class_factory or sqlalchemy.ext.declarative.declarative_base
    Base = cls or (class_factory)()
    target_metadata = []
    apps = apps or getattr(settings, 'INSTALLED_APPS', [])
    if apps:
        for app_qualname in apps:
            orm_qualname = "%s.orm" % app_qualname
            try:
                orm = importlib.import_module(orm_qualname)
            except ImportError as e:
                # Fail silently - assume that the installed application does
                # not declare an ORM. We do want to catch other ImportError
                # exceptions tho. This is quicky flaky but it works for now.
                if e.args[0] != ("No module named '%s'" %  orm_qualname):
                    raise
                continue
            if isinstance(orm.metadata, MetaData):
                target_metadata.append(orm.metadata)
            elif isinstance(orm.metadata, (list, tuple)):
                target_metadata.extend(orm.metadata)
            else:
                raise TypeError(
                    "%s.metadata is of invalid type." % orm_qualname)

    target_metadata.append(Base.metadata)
    return Base, target_metadata


def destroy_engines(loop=None):
    """Destroys all known engines."""
    for name in list(dict.keys(ENGINES)):
        engine = ENGINES.pop(name)
        engine.dispose()

    for name in list(dict.keys(ASYNC_ENGINES)):
        engine = ASYNC_ENGINES.pop(name)
        asyncio.run(engine.dispose())\
            if loop is None\
            else loop.run_until_complete(engine.dispose())


def create_engine(name='self', *args, **kwargs):
    """Create an :class:`sqlalchemy.engine.Engine` instance used the named
    connection `name`.
    """
    global ENGINES
    global ASYNC_ENGINES

    databases = kwargs.pop('databases', None) or DATABASES
    use_async = kwargs.get('use_async', False)

    engine_factory = create_async_engine
    if not use_async:
        engine_factory = sqlalchemy.create_engine
    engines = ENGINES if not use_async else ASYNC_ENGINES
    if name not in engines:
        engines[name] = databases[name]\
            .as_engine(engine_factory, *args, **kwargs)
    return engines[name]


def get_dsn(name='self', use_async=False):
    """Return the Data Source Name (DSN) for the named connection."""
    global DATABASES
    return DATABASES[name].as_dsn(use_async=use_async)


def session_factory(name='self', *args, **kwargs):
    """Create a new session factory for the named connection `name`."""
    return sessionmaker(bind=create_engine(name), *args, **kwargs)


def async_session(name='self'):
    """Create a new asyncronous session for the named connection `name`."""
    engine = create_engine(name, use_async=True)
    return AsyncSession(engine)



# pylint: skip-file
import os
import unittest

from .. import awaitservices
from .. import destroy_engines
from ..conf import load_config


class AwaitServicesTestCase(unittest.TestCase):

    def tearDown(self):
        destroy_engines()

    def test_happy_sqlite(self):
        databases = load_config({
            'DB_ENGINE': 'sqlite',
            'DB_NAME': ":memory:",
        })
        awaitservices(max_retries=1, interval=0.1, databases=databases)

    def test_non_resolvable_hostname_postgresql(self):
        databases = load_config({
            'DB_ENGINE': 'postgresql',
            'DB_NAME': "rdbms",
            'DB_USERNAME': "rdbms",
            'DB_PASSWORD': "rdbms",
            'DB_HOST': bytes.hex(os.urandom(16))
        })
        with self.assertRaises(RuntimeError):
            awaitservices(max_retries=1, interval=0.1, databases=databases)



# pylint: skip-file
import asyncio

import unittest
import unittest.mock

from unimatrix.lib.datastructures import ImmutableDTO
from unimatrix.ext.model.exc import ValidationError

from .. import query


class NamedQueryTestCase(unittest.TestCase):

    class query_class(query.NamedQuery):
        foo = query.String(required=True)

        def build(self):
            return None

    def test_instantiate_with_valid_parameters(self):
        q = self.query_class(foo='bar')

    def test_instantiate_with_invalid_parameters(self):
        with self.assertRaises(ValidationError):
            q = self.query_class(foo=1)

    def test_parameter_descriptor(self):
        q = self.query_class(foo='bar')
        self.assertEqual(q.foo, 'bar')

    def test_cannot_create_class_with_forbidden_attribute(self):
        with self.assertRaises(ValueError):
            type('Foo', (query.NamedQuery,), {
                'build': query.String()
            })

    def test_run_does_emit_query(self):
        exc = Exception()
        q = self.query_class(foo='bar')
        f = q.run(ImmutableDTO(execute=unittest.mock.AsyncMock(side_effect=exc)))
        try:
            result = asyncio.run(f)
        except Exception  as e:
            self.assertEqual(e, exc)



# pylint: skip-file
import unittest

from .. import create_engine
from .. import destroy_engines
from .. import ENGINES
from .. import ASYNC_ENGINES
from ..conf import load_config


class CreateEngineTestCase(unittest.TestCase):

    def test_engine_is_created(self):
        engine = create_engine('self',
            databases=load_config({'DB_ENGINE': 'sqlite', 'DB_NAME': ':memory:'}))
        self.assertEqual(engine.dialect.name, 'sqlite')

    def test_registry_is_updated(self):
        destroy_engines()
        self.assertEqual(len(ENGINES), 0)
        engine = create_engine('self',
            databases=load_config({'DB_ENGINE': 'sqlite', 'DB_NAME': ':memory:'}))
        self.assertEqual(len(ASYNC_ENGINES), 0)
        self.assertEqual(len(ENGINES), 1)

    def test_async_registry_is_updated(self):
        destroy_engines()
        self.assertEqual(len(ASYNC_ENGINES), 0)
        engine = create_engine('self',
            use_async=True,
            databases=load_config({'DB_ENGINE': 'sqlite', 'DB_NAME': ':memory:'}))
        self.assertEqual(len(ASYNC_ENGINES), 1)
        self.assertEqual(len(ENGINES), 0)
